#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

struct dns_header {
  uint16_t id;
  uint16_t flags;
#define DNS_FLAG_QR (1 << 15)
#define DNS_FLAG_AA (1 << 10)
#define DNS_FLAG_TC (1 <<  9)
#define DNS_FLAG_RD (1 <<  8)
#define DNS_FLAG_RA (1 <<  7)
#define DNS_FLAG_AD (1 <<  5)
#define DNS_FLAG_CD (1 <<  4)
#define DNS_CODE_OPCODE_SHIFT  11
#define DNS_CODE_OPCODE_MASK  0xf
#define DNS_CODE_RCODE_SHIFT    0
#define DNS_CODE_RCODE_MASK   0xf
  uint16_t qdcount;
  uint16_t ancount;
  uint16_t nscount;
  uint16_t arcount;
};
#define DNS_HEADER_SIZE 12

struct dns_question {
  uint16_t type;
  uint16_t class;
};
#define DNS_QUESTION_SIZE 4

struct dns_rr {
  uint16_t type;
  uint16_t class;
  uint32_t ttl;
  uint16_t rdlength;
};
#define DNS_RR_SIZE 10

struct resource_record {
  struct resource_record *next;
  int type;
  char *hostname;
  union {
    in_addr_t ipaddr;
    struct in6_addr ip6addr;
    char domainname[32];
  } rdata;
};

struct client {
  struct client *next;
  in_addr_t ipaddr;
  int port;
  int type;
  char name[1];
};

static struct client *clients = NULL;

static struct client *client_add(in_addr_t ipaddr, int port,
				 char *name, int type)
{
  struct client *client;
  client = malloc(sizeof(*client) + strlen(name));

  client->ipaddr = ipaddr;
  client->port = port;
  strcpy(client->name, name);
  client->type = type;

  client->next = clients;
  clients = client;

  return client;
}

static struct client *client_delete(char *name, int type)
{
  struct client *client, **p;

  for (p = &clients; *p; p = &((*p)->next)) {
    if (!strcmp((*p)->name, name) && ((*p)->type == type)) {
      client = *p;
      *p = (*p)->next;
      client->next = NULL;
      return client;
    }
  }
  return NULL;
}

static unsigned char *write_name(struct dns_header *dnshdr, unsigned char *p,
				 int size, char *name)
{
  char *np = name, *pp;

  while (1) {
    pp = strchr(np, '.');
    if (!pp)
      pp = np + strlen(np);
    if (size < 1 + (pp - np) + 1)
      return NULL;
    *(p++) = pp - np;
    size--;
    memcpy(p, np, pp - np);
    p += pp - np;
    size -= pp - np;
    if (*pp == '\0')
      break;
    np = pp + 1;
  }
  *(p++) = 0x00;

  return p;
}

static int make_dns_header(unsigned char *buffer, int size, int id, int flags,
			   int qdcount, int ancount, int nscount, int arcount)
{
  struct dns_header header;

  if (size < DNS_HEADER_SIZE)
    return -1;

  header.id = htons(id);
  header.flags = htons(flags);

  header.qdcount = htons(qdcount);
  header.ancount = htons(ancount);
  header.nscount = htons(nscount);
  header.arcount = htons(arcount);

  memcpy(buffer, &header, DNS_HEADER_SIZE);

  return DNS_HEADER_SIZE;
}

static int make_dns_question(struct dns_header *dnshdr, unsigned char *buffer,
			     int size, char *name, int type, int class)
{
  struct dns_question question;
  unsigned char *p = buffer;

  p = write_name(dnshdr, p, size, name);
  if (!p)
    return -1;
  size -= p - buffer;

  if (size < DNS_QUESTION_SIZE)
    return -1;

  question.type  = htons(type);
  question.class = htons(class);

  memcpy(p, &question, DNS_QUESTION_SIZE);
  p += DNS_QUESTION_SIZE;

  return p - buffer;
}

static int make_dns_rr(struct dns_header *dnshdr, unsigned char *buffer,
		       int size, char *name, int type, int class,
		       int ttl, void *rdata)
{
  struct dns_rr rr;
  unsigned char *p = buffer, *rrp;
  in_addr_t inaddr;

  p = write_name(dnshdr, p, size, name);
  if (!p)
    return -1;
  size -= p - buffer;

  if (size < DNS_RR_SIZE)
    return -1;

  rrp = p;
  p += DNS_RR_SIZE;
  size -= DNS_RR_SIZE;

  switch (type) {
  case  1: /* A (IPv4 address) */
    memcpy(&inaddr, rdata, sizeof(inaddr));
    inaddr = htonl(inaddr);
    memcpy(p, &inaddr, sizeof(inaddr));
    p += sizeof(inaddr);
    break;
  case  2: /* NS */
  case 12: /* PTR */
    p = write_name(dnshdr, p, size, rdata);
    break;
  case 15: /* MX */
    memcpy(p, "\x00\x0a", 2);
    p = write_name(dnshdr, p + 2, size, rdata);
    break;
  case 28: /* IPv6 */
    memcpy(p, rdata, sizeof(struct in6_addr));
    p += sizeof(struct in6_addr);
    break;
  default:
    memcpy(p, rdata, strlen(rdata));
    p += strlen(rdata) + 1;
    break;
  }

  rr.type     = htons(type);
  rr.class    = htons(class);
  rr.ttl      = htonl(ttl);
  rr.rdlength = htons(p - (rrp + DNS_RR_SIZE));
  memcpy(rrp, &rr, DNS_RR_SIZE);

  return p - buffer;
}

static unsigned char *read_name(struct dns_header *dnshdr, unsigned char *p,
				char *name, int size)
{
  int s, count = 0;
  char *np = name;
  unsigned char *end = NULL;

  while ((s = *(p++)) != 0) {
    if ((s & 0xc0) == 0xc0) {
      s = ((s & ~0xc0) << 8) + *(p++);
      if (!end) end = p;
      p = (unsigned char *)dnshdr + s;
      continue;
    }
    if (np - name + s + (count ? 1 : 0) + 1 > size)
      return NULL;
    if (count++) *(np++) = '.';
    memcpy(np, p, s);
    np += s;
    p += s;
  }
  *np = '\0';
  if (!end) end = p;

  return end;
}

static unsigned char *analyze_dns_header(struct dns_header *dnshdr,
					 unsigned char *p,
					 struct dns_header *header)
{
  header->id      = ntohs(dnshdr->id);
  header->flags   = ntohs(dnshdr->flags);
  header->qdcount = ntohs(dnshdr->qdcount);
  header->ancount = ntohs(dnshdr->ancount);
  header->nscount = ntohs(dnshdr->nscount);
  header->arcount = ntohs(dnshdr->arcount);

  p += DNS_HEADER_SIZE;

  printf("DNS:\n");
  printf("\tID\t: %04x\n", header->id);

  printf("\tFLAGS\t:");
  if (header->flags & DNS_FLAG_QR) printf(" QR");
  if (header->flags & DNS_FLAG_AA) printf(" AA");
  if (header->flags & DNS_FLAG_TC) printf(" TC");
  if (header->flags & DNS_FLAG_RD) printf(" RD");
  if (header->flags & DNS_FLAG_RA) printf(" RA");
  if (header->flags & DNS_FLAG_AD) printf(" AD");
  if (header->flags & DNS_FLAG_CD) printf(" CD");
  printf("\n");

  printf("\tOPCODE\t: %d\n",
	 (header->flags >> DNS_CODE_OPCODE_SHIFT) & DNS_CODE_OPCODE_MASK);
  printf("\tRCODE\t: %d\n",
	 (header->flags >> DNS_CODE_RCODE_SHIFT) & DNS_CODE_RCODE_MASK);

  printf("\tQDCOUNT\t: %d\n", header->qdcount);
  printf("\tANCOUNT\t: %d\n", header->ancount);
  printf("\tNSCOUNT\t: %d\n", header->nscount);
  printf("\tARCOUNT\t: %d\n", header->arcount);

  return p;
}

static unsigned char *analyze_dns_question(struct dns_header *dnshdr,
					   unsigned char *p,
					   char *name, int size,
					   struct dns_question *question)
{
  p = read_name(dnshdr, p, name, size);
  if (!p) return NULL;

  memcpy(question, p, DNS_QUESTION_SIZE);
  question->type  = ntohs(question->type);
  question->class = ntohs(question->class);

  p += DNS_QUESTION_SIZE;

  printf("\t\tname\t: %s\n", name);
  printf("\t\ttype\t: %d\n", question->type);
  printf("\t\tclass\t: %d\n", question->class);

  return p;
}

static int analyze_dns(struct dns_header *dnshdr, struct dns_header *header,
		       char *name, int size, struct dns_question *question)
{
  unsigned char *p = (unsigned char *)dnshdr;

  p = analyze_dns_header(dnshdr, p, header);
  if (!p) return -1;

  if (header->qdcount < 1)
    return 0;

  p = analyze_dns_question(dnshdr, p, name, size, question);
  if (!p) return -1;

  return 1;
}

int main(int argc, char *argv[])
{
  int sockt, port, upper_port, size;
  in_addr_t server_ipaddr, upper_ipaddr;
  char *server_hostname;
  unsigned char buffer[4096], *p;
  struct sockaddr_in addr;
  socklen_t addrlen;
  struct dns_header header;
  char question_name[256];
  struct dns_question question;
  struct resource_record *rr = NULL, *rrp;
  struct client *client;

  argc--; argv++;

  upper_ipaddr = ntohl(inet_addr(argv[0]));
  argc--; argv++;
  upper_port = atoi(argv[0]);
  argc--; argv++;

  port = atoi(argv[0]);
  argc--; argv++;

  server_hostname = argv[0];
  argc--; argv++;
  server_ipaddr = ntohl(inet_addr(argv[0]));
  argc--; argv++;

  while (argc >= 3) {
    rrp = malloc(sizeof(*rrp));
    memset(rrp, 0, sizeof(*rrp));
    if      (!strcmp(argv[0], "A"   )) rrp->type =  1;
    else if (!strcmp(argv[0], "PTR" )) rrp->type = 12;
    else if (!strcmp(argv[0], "MX"  )) rrp->type = 15;
    else if (!strcmp(argv[0], "AAAA")) rrp->type = 28;
    argc--; argv++;

    rrp->hostname = strdup(argv[0]);
    argc--; argv++;
    switch (rrp->type) {
    case  1: /* A (IPv4 address) */
      rrp->rdata.ipaddr = ntohl(inet_addr(argv[0]));
      break;
    case 28: /* IPv6 */
      inet_pton(AF_INET6, argv[0], &rrp->rdata.ip6addr);
      break;
    case 12: /* PTR */
    case 15: /* MX */
    default:
      strcpy(rrp->rdata.domainname, argv[0]);
      break;
    }
    argc--; argv++;

    rrp->next = rr;
    rr = rrp;
  }

  sockt = socket(PF_INET, SOCK_DGRAM, 0);

  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = htons(port);
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  addrlen = sizeof(addr);

  if (bind(sockt, (struct sockaddr *)&addr, addrlen) < 0) {
    fprintf(stderr, "Cannot bind.\n");
    exit(1);
  }

  while (1) {
    memset(&addr, 0, sizeof(addr));
    addrlen = sizeof(addr);
    size = recvfrom(sockt, buffer, sizeof(buffer), 0,
		    (struct sockaddr *)&addr, &addrlen);
    if (size < 0) {
      fprintf(stderr, "Cannot recv.\n");
      exit(1);
    }
    if (size == sizeof(buffer)) {
      fprintf(stderr, "Out of buffer.\n");
      exit(1);
    }

    if (analyze_dns((struct dns_header *)buffer, &header,
		    question_name, sizeof(question_name), &question) != 1)
      continue;

    if (header.flags & DNS_FLAG_QR) {
      client = client_delete(question_name, question.type);
      if (client == NULL)
	continue;

      p = buffer + size;

      memset(&addr, 0, sizeof(addr));
      addr.sin_family = AF_INET;
      addr.sin_port = htons(client->port);
      addr.sin_addr.s_addr = htonl(client->ipaddr);
      addrlen = sizeof(addr);

      goto send;
    }

    for (rrp = rr; rrp; rrp = rrp->next) {
      if ((question.type == rrp->type) &&
	  !strcmp(question_name, rrp->hostname))
	break;
    }
    if (rrp == NULL) {
      client_add(ntohl(addr.sin_addr.s_addr), ntohs(addr.sin_port),
		 question_name, question.type);

      size = make_dns_header(buffer, sizeof(buffer), header.id,
			     DNS_FLAG_RD, 1, 0, 0, 0);
      if (size < 0) return -1;
      p = buffer + size;

      size = make_dns_question((struct dns_header *)buffer, p,
			       sizeof(buffer) - (p - buffer),
			       question_name, question.type, question.class);
      if (size < 0) return -1;
      p += size;

      memset(&addr, 0, sizeof(addr));
      addr.sin_family = AF_INET;
      addr.sin_port = htons(upper_port);
      addr.sin_addr.s_addr = htonl(upper_ipaddr);
      addrlen = sizeof(addr);

      goto send;
    }

    size = make_dns_header(buffer, sizeof(buffer), header.id,
			   DNS_FLAG_QR|DNS_FLAG_AA|DNS_FLAG_RD|DNS_FLAG_RA,
			   1, 1, 1, 1);
    if (size < 0) return -1;
    p = buffer + size;

    size = make_dns_question((struct dns_header *)buffer, p,
			     sizeof(buffer) - (p - buffer), question_name,
			     question.type, question.class);
    if (size < 0) return -1;
    p += size;

    size = make_dns_rr((struct dns_header *)buffer, p,
		       sizeof(buffer) - (p - buffer), rrp->hostname,
		       question.type, question.class, 86400, &rrp->rdata);
    if (size < 0) return -1;
    p += size;

    size = make_dns_rr((struct dns_header *)buffer, p,
		       sizeof(buffer) - (p - buffer), rrp->hostname,
		       2, 1, 86400, server_hostname);
    if (size < 0) return -1;
    p += size;

    size = make_dns_rr((struct dns_header *)buffer, p,
		       sizeof(buffer) - (p - buffer), server_hostname,
		       1, 1, 86400, &server_ipaddr);
    if (size < 0) return -1;
    p += size;

send:
    size = p - buffer;
    if (sendto(sockt, buffer, size, 0, (struct sockaddr *)&addr, addrlen)
	!= size) {
      fprintf(stderr, "Cannot send.\n");
      exit(1);
    }
  }

  return 0;
}
